#ifndef ENIGMATIC_MONITOR_NETWORK_INTERFACES_H
#define ENIGMATIC_MONITOR_NETWORK_INTERFACES_H

#include "Enigmatic.h"

Eina_Bool
enigmatic_monitor_network_interfaces(Enigmatic *enigmatic, Eina_Hash **cache_hash);

#endif
