#ifndef ENIGMATIC_MONITOR_H
#define ENIGMATIC_MONITOR_H

#include "Enigmatic.h"
#include "cores.h"
#include "memory.h"
#include "batteries.h"
#include "power.h"
#include "sensors.h"
#include "file_systems.h"
#include "network_interfaces.h"
#include "processes.h"

#endif
