#ifndef ENIGMATIC_MONITOR_FILE_SYSTEMS_H
#define ENIGMATIC_MONITOR_FILE_SYSTEMS_H

#include "Enigmatic.h"

Eina_Bool
enigmatic_monitor_file_systems(Enigmatic *enigmatic, Eina_Hash **cache_hash);

#endif
